import bbi
import cooler
import functools as ft
import logging
import numpy as np
import pandas as pd
import re
import time

TILE_SIZE = 1024

logger = logging.getLogger(__name__)

def get_quadtree_depth(chromsizes):
    tile_size_bp = TILE_SIZE
    min_tile_cover = np.ceil(sum(chromsizes) / tile_size_bp)
    return int(np.ceil(np.log2(min_tile_cover)))

def get_zoom_resolutions(chromsizes):
    return [2**x for x in range(get_quadtree_depth(chromsizes) + 1)][::-1]

def natsort_key(s, _NS_REGEX=re.compile(r'(\d+)', re.U)):
        return tuple([int(x) if x.isdigit() else x
                             for x in _NS_REGEX.split(s) if x])

def natcmp(x, y):
    if x.find('_') >= 0:
        x_parts = x.split('_')
        if y.find('_') >= 0:
            # chr_1 vs chr_2
            y_parts = y.split('_')

            return natcmp(x_parts[1], y_parts[1])
        else:
            # chr_1 vs chr1
            # chr1 comes first
            return 1
    if y.find('_') >= 0:
        # chr1 vs chr_1
        # y comes second
        return -1

    _NS_REGEX=re.compile(r'(\d+)', re.U)
    x_parts = tuple([int(a) if a.isdigit() else a
                             for a in _NS_REGEX.split(x) if a])
    y_parts = tuple([int(a) if a.isdigit() else a
                             for a in _NS_REGEX.split(y) if a])

    if x == 'chrM':
        # chrM goes at the end of the non alternate contigs
        return 1
    if y == 'chrM':
        return -1

    if x_parts < y_parts:
        return -1
    elif y_parts > x_parts:
        return 1
    else:
        return 0

def natsorted(iterable):
    return sorted(iterable, key=ft.cmp_to_key(natcmp))

def get_chromsizes(bwpath):
    """
    TODO: replace this with negspy
    
    Also, return NaNs from any missing chromosomes in bbi.fetch
    
    """
    chromsizes = bbi.chromsizes(bwpath)
    chromosomes = natsorted(chromsizes.keys())
    print("chromosomes", chromosomes)
    chrom_series = pd.Series(chromsizes)[chromosomes]
    return chrom_series

def abs2genomic(chromsizes, start_pos, end_pos):
    abs_chrom_offsets = np.r_[0, np.cumsum(chromsizes.values)]
    cid_lo, cid_hi = np.searchsorted(abs_chrom_offsets,
                                     [start_pos, end_pos],
                                     side='right') - 1
    rel_pos_lo = start_pos - abs_chrom_offsets[cid_lo]
    rel_pos_hi = end_pos - abs_chrom_offsets[cid_hi]
    start = rel_pos_lo
    for cid in range(cid_lo, cid_hi):
        yield cid, start, chromsizes[cid]
        start = 0
    yield cid_hi, start, rel_pos_hi

def tileset_info(bwpath):
    '''
    Get the tileset info for a bigWig file

    Parameters
    ----------
    bwpath: string
        The path to the bigwig file from which to retrieve data

    Returns
    -------
    tileset_info: {'min_pos': [], 
                    'max_pos': [], 
                    'tile_size': 1024, 
                    'max_zoom': 7
                    }
    '''
    TILE_SIZE = 1024
    chromsizes = bbi.chromsizes(bwpath)
    chromosomes = cooler.util.natsorted(chromsizes.keys())
    chromsizes = pd.Series(chromsizes)[chromosomes]
    min_tile_cover = np.ceil(sum(chromsizes) / TILE_SIZE)
    max_zoom = int(np.ceil(np.log2(min_tile_cover)))
    tileset_info = {
        'min_pos': [0],
        'max_pos': [TILE_SIZE * 2 ** max_zoom],
        'max_width': TILE_SIZE * 2 ** max_zoom,
        'tile_size': TILE_SIZE,
        'max_zoom': max_zoom
    }
    return tileset_info

def get_bigwig_tile(bwpath, zoom_level, start_pos, end_pos, chromsizes=None):
    t1 = time.time()
    if chromsizes is None:
        print("here")
        chromsizes = get_chromsizes(bwpath)
    print('chromsizes:', chromsizes)
    t2 = time.time()

    # print("chromosomes time:", t2 - t1)
    resolutions = get_zoom_resolutions(chromsizes)
    binsize = resolutions[zoom_level]
   
    arrays = []
    for cid, start, end in abs2genomic(chromsizes, start_pos, end_pos):
        n_bins = int(np.ceil((end - start) / binsize))
        try:
            chrom = chromsizes.index[cid]
            clen = chromsizes.values[cid]

            t1 = time.time()
            #print("fetching:", chrom, start, end, n_bins);
            x = bbi.fetch(bwpath, chrom, start, end,
                          bins=n_bins, missing=np.nan)
            t2 = time.time()

            if t2 - t1 > 0.5:
                print("fetching:", chrom, start, end, n_bins, "fetched time: {:.2f}".format(time.time() - t1))

            # drop the very last bin if it is smaller than the binsize
            if end == clen and clen % binsize != 0:
                x = x[:-1]
        except IndexError:
            # beyond the range of the available chromosomes
            # probably means we've requested a range of absolute
            # coordinates that stretch beyond the end of the genome
            x = np.zeros(n_bins)

        arrays.append(x)

    return np.concatenate(arrays)


def tiles(bwpath, tile_ids, chromsizes_map={}):
    '''
    Generate tiles from a bigwig file.

    Parameters
    ----------
    tileset: tilesets.models.Tileset object
        The tileset that the tile ids should be retrieved from
    tile_ids: [str,...]
        A list of tile_ids (e.g. xyx.0.0) identifying the tiles
        to be retrieved
    chromsizes_map: {uid: []}
        A set of chromsizes listings corresponding to the parameters of the 
        tile_ids

    Returns
    -------
    tile_list: [(tile_id, tile_data),...]
        A list of tile_id, tile_data tuples
    '''
    import base64
    TILE_SIZE = 1024
    generated_tiles = []
    for tile_id in tile_ids:
        tile_option_parts = tile_id.split('|')[1:]
        tile_no_options = tile_id.split('|')[0]
        tile_id_parts = tile_no_options.split('.')
        tile_position = list(map(int, tile_id_parts[1:3]))

        tile_options = dict([o.split(':') for o in tile_option_parts])

        chromsizes_id = None
        if 'cos' in tile_options:
            chromsizes_id = tile_options['cos']

        print('chromsizes_id:', chromsizes_id)

        if chromsizes_id in chromsizes_map:
            chromsizes = chromsizes_map[chromsizes_id]
        else:
            chromsizes = None

        zoom_level = tile_position[0]
        tile_pos = tile_position[1]
        
        # this doesn't combine multiple consequetive ids, which
        # would speed things up
        if chromsizes is None:
            chromsizes = get_chromsizes(bwpath)

        print('chromsizes:', chromsizes)
        max_depth = get_quadtree_depth(chromsizes)
        tile_size = TILE_SIZE * 2 ** (max_depth - zoom_level)
        start_pos = tile_pos * tile_size
        end_pos = start_pos + tile_size
        dense = get_bigwig_tile(bwpath, zoom_level, start_pos, end_pos, chromsizes)

        if len(dense):
            max_dense = max(dense)
            min_dense = min(dense)
        else:
            max_dense = 0
            min_dense = 0

        min_f16 = np.finfo('float16').min
        max_f16 = np.finfo('float16').max

        has_nan = len([d for d in dense if np.isnan(d)]) > 0

        if (
            not has_nan and
            max_dense > min_f16 and max_dense < max_f16 and
            min_dense > min_f16 and min_dense < max_f16
        ):
            tile_value = {
                'dense': base64.b64encode(dense.astype('float16')).decode('utf-8'),
                'dtype': 'float16'
            }
        else:
            tile_value = {
                'dense': base64.b64encode(dense.astype('float32')).decode('utf-8'),
                'dtype': 'float32'
            }

        generated_tiles += [(tile_id, tile_value)]
    return generated_tiles

def chromsizes(filename):
    '''
    Get a list of chromosome sizes from this [presumably] bigwig
    file.

    Parameters:
    -----------
    filename: string
        The filename of the bigwig file

    Returns
    -------
    chromsizes: [(name:string, size:int), ...]
        An ordered list of chromosome names and sizes
    '''
    try:
        chrom_series = get_chromsizes(filename)
        data = []
        for chrom, size in chrom_series.iteritems():
            data.append([chrom, size])
        return data
    except Exception as ex:
        logger.error(ex)
        raise Exception('Error loading chromsizes from bigwig file: {}'.format(ex))



import json

def tileset_info(filename):
    '''
    Return a tileset info for this time interval
    '''
    with open(filename, 'r') as f:
        data = json.load(f)

        return {
            'min_pos': [0],
            'max_pos': [data['len']],
            'max_width': data['len'],
            'start_value': data['start'],
            'end_value': data['end']
        }



import base64
import os
import sqlite3

def get_tileset_info(tileset):
    if not os.path.isfile(tileset):
        return {
            'error': 'Tileset info is not available!'
        }

    db = sqlite3.connect(tileset)

    res = db.execute('SELECT * FROM tileset_info').fetchone()

    o = {
        'tile_size': res[5],
        'max_zoom': res[6],
        'max_size': res[7],
    }

    try:
        o['width'] = res[8]
        o['height'] = res[9]
    except IndexError:
        pass

    try:
        o['dtype'] = res[10]
    except IndexError:
        pass

    return o

def get_tiles(filename, tile_ids, raw):
    '''
    Generate tiles from a imtiles file.
    Parameters
    ----------
    tileset: tilesets.models.Tileset object
        The tileset that the tile ids should be retrieved from
    tile_ids: [str,...]
        A list of tile_ids (e.g. xyx.0.0.1) identifying the tiles
        to be retrieved
    Returns
    -------
    generated_tiles: [(tile_id, tile_data),...]
        A list of tile_id, tile_data tuples
    '''
    # Connect to SQLite db
    db = sqlite3.connect(filename)

    generate_tiles = []

    generate_image = raw and len(tile_ids)

    for tile_id in tile_ids:
        id = tile_id[tile_id.find('.') + 1:].split('.')

        sql = 'SELECT image FROM tiles WHERE z = :z AND y = :y AND x = :x'
        param = {'z': int(id[0]), 'y': int(id[1]), 'x': int(id[2])}
        res = db.execute(sql, param).fetchone()

        if res:
            image_blob = res[0]

            if generate_image:
                tile_data = {
                    'image': image_blob,
                }
            else:
                tile_data = {
                    'dense': base64.b64encode(image_blob).decode('latin-1'),
                }

            generate_tiles.append((tile_id, tile_data))

    return generate_tiles



import os.path as op

def partition_by_adjacent_tiles(tile_ids, dimension=2):
    '''
    Partition a set of tile ids into sets of adjacent tiles

    Parameters
    ----------
    tile_ids: [str,...]
        A list of tile_ids (e.g. xyx.0.0.1) identifying the tiles
        to be retrieved
    dimension: int
        The dimensionality of the tiles

    Returns
    -------
    tile_lists: [tile_ids, tile_ids]
        A list of tile lists, all of which have tiles that
        are within 1 position of another tile in the list
    '''
    tile_id_lists = []

    for tile_id in sorted(tile_ids, key=lambda x: [int(p) for p in x.split('.')[2:2+dimension]]):
        tile_id_parts = tile_id.split('.')

        # exclude the zoom level in the position
        # because the tiles should already have been partitioned
        # by zoom level
        tile_position = list(map(int, tile_id_parts[2:4]))

        added = False

        for tile_id_list in tile_id_lists:
            # iterate over each group of adjacent tiles
            has_close_tile = False

            for ct_tile_id in tile_id_list:
                ct_tile_id_parts = ct_tile_id.split('.')
                ct_tile_position = list(map(int, ct_tile_id_parts[2:2+dimension]))
                far_apart = False

                # iterate over each dimension and see if this tile is close
                for p1,p2 in zip(tile_position, ct_tile_position):
                    if abs(int(p1) - int(p2)) > 1:
                        # too far apart can't be part of the same group
                        far_apart = True

                if not far_apart:
                    # no position was too far
                    tile_id_list += [tile_id]
                    added = True
                    break

            if added:
                break
        if not added:
            tile_id_lists += [[tile_id]]

    return tile_id_lists

def infer_filetype(filename):
    _,ext = op.splitext(filename)

    if ext.lower() == '.bw' or ext.lower() == '.bigwig':
        return 'bigwig'
    elif ext.lower() == '.mcool' or ext.lower() == '.cool':
        return 'cooler'
    elif ext.lower() == '.htime':
        return 'time-interval-json'
    elif ext.lower() == '.hitile':
        return 'hitile'
    elif ext.lower() == '.beddb':
        return 'beddb'
    elif ext.lower() == '.mv5':
        return 'multivec'

    return None

def infer_datatype(filetype):
    if filetype == 'cooler':
        return 'matrix'
    if filetype == 'bigwig':
        return 'vector'
    if filetype == 'time-interval-json':
        return 'time-interval'
    if filetype == 'hitile':
        return 'vector'
    if filetype == 'beddb':
        return 'bedlike'



from setuptools import setup

setup(
    name='hgtiles',
    version='0.2.9',
    description='Tile loading for higlass-server',
    author='Peter Kerpedjiev',
    author_email='pkerpedjiev@gmail.com',
    install_requires=[
        'pandas>=0.19',
    ],
    url='',
    packages=['hgtiles'],
)



import collections as col
import math
import sqlite3

def tileset_info(db_file):
    conn = sqlite3.connect(db_file)
    c = conn.cursor()

    row = c.execute("SELECT * from tileset_info").fetchone();
    if row is not None and len(row) == 9:
        header = row[8]
    else:
        header = ""

    ts_info = {
            'zoom_step': row[0],
            'max_length': row[1],
            'assembly': row[2],
            'chrom_names': row[3],
            'chrom_sizes': row[4],
            'tile_size': row[5],
            'max_zoom': row[6],
            'max_width': row[7],
            "min_pos": [1],
            "max_pos": [row[1]],
            "header": header
            }
    conn.close()

    return ts_info

def tiles(filepath, tile_ids):
    '''
    Generate tiles from this dataset.

    Parameters
    ----------
    filepath: str
        The filename of the sqlite db file
    tile_ids: [str...]
        A list of tile ids of the form

    Returns
    -------
    tiles: [(tile_id, tile_value),...]
        A list of values indexed by the tile position
    '''
    to_return = []

    for  tile_id in tile_ids:
        tile_option_parts = tile_id.split('|')[1:]
        tile_no_options = tile_id.split('|')[0]
        parts = tile_no_options.split('.')

        uuid = parts[0]
        zoom = int(parts[1])
        xpos = int(parts[2])

        extra_zoom = 0
        new_rows = {}
        new_rows = []

        for j in range(2 ** extra_zoom):
            # the old rows are indexed by the higher
            # resolution tile numbers
            higher_xpos = 2 ** extra_zoom * xpos + j
            old_rows = get_1D_tiles(filepath, 
                    zoom + extra_zoom, 
                    higher_xpos)
            new_rows += old_rows
                
        print("new_rows length", len(new_rows))
        to_return += [(tile_id, new_rows)]

    return to_return

def get_1D_tiles(db_file, zoom, tile_x_pos, num_tiles=1):
    '''
    Retrieve a contiguous set of tiles from a db tile file.

    Parameters
    ----------
    db_file: str
        The filename of the sqlite db file
    zoom: int
        The zoom level
    tile_x_pos: int
        The position of the first tile
    num_tiles: int
        The number of tiles to retrieve

    Returns
    -------
    tiles: {pos: tile_value}
        A set of tiles, indexed by position
    '''
    ts_info = tileset_info(db_file)
    conn = sqlite3.connect(db_file)

    c = conn.cursor()

    tile_width = ts_info['max_width'] / 2 ** zoom

    tile_start_pos = tile_width * tile_x_pos
    tile_end_pos = tile_start_pos + num_tiles * tile_width

    query = '''
    SELECT startPos, endPos, chrOffset, importance, fields, uid
    FROM intervals,position_index
    WHERE
        intervals.id=position_index.id AND
        zoomLevel <= {} AND
        rEndPos >= {} AND
        rStartPos <= {}
    '''.format(zoom, tile_start_pos, tile_end_pos)

    rows = c.execute(query).fetchall()

    new_rows = []

    for r in rows:
        try:
            uid = r[5].decode('utf-8')
        except AttributeError:
            uid = r[5]

        tile_pos = (
            tile_x_pos + math.floor((r[0] - tile_start_pos) / tile_width)
        )

        x_start = r[0]
        x_end = r[1]

        for i in range(tile_x_pos, tile_x_pos + num_tiles):
            tile_x_start = i * tile_width
            tile_x_end = (i+1) * tile_width
            tile_pos = i

            if x_start < tile_x_end and x_end >= tile_x_start:
                new_rows += [
                    # add the position offset to the returned values
                    {'xStart': r[0],
                     'xEnd': r[1],
                     'chrOffset': r[2],
                     'importance': r[3],
                     'uid': uid,
                     'fields': r[4].split('\t')}]
    conn.close()

    return new_rows

def list_items(db_file, start, end, max_entries=None):
    '''
    List the entries between start and end

    Parameters
    ----------
    db_file: str
        The filename of the sqlite db file
    start_pos: int
        The start position from where to retrieve data
    end_pos: int
        The end position to get data
    max_entries: int
        The maximum number of results to return
    '''

    conn = sqlite3.connect(db_file)

    c = conn.cursor()

    # some large number because we want to extract all entries
    zoom = 100000

    query = '''
    SELECT startPos, endPos, chrOffset, importance, fields, uid
    FROM intervals,position_index
    WHERE
        intervals.id=position_index.id AND
        zoomLevel <= {} AND
        rEndPos >= {} AND
        rStartPos <= {}
    '''.format(zoom, start, end)

    if max_entries is not None:
        query += ' LIMIT {}'.format(max_entries)

    rows = c.execute(query).fetchall()

    new_rows = []

    for r in rows:
        try:
            uid = r[5].decode('utf-8')
        except AttributeError:
            uid = r[5]

        new_rows += [
            # add the position offset to the returned values
            {'xStart': r[0],
             'xEnd': r[1],
             'chrOffset': r[2],
             'importance': r[3],
             'uid': uid,
             'fields': r[4].split('\t')}]
    conn.close()

    return new_rows
